@@ -2,7 +2,7 @@
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
-#      http://www.apache.org/licenses/LICENSE-2.0
+#      https://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
@@ -10,93 +10,26 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
-import json
 import weakref
 
-import requests
 from requests.adapters import BaseAdapter
 import six
 from six.moves.urllib import parse as urlparse
 
 from requests_mock import exceptions
-from requests_mock import response
+from requests_mock.request import _RequestObjectProxy
+from requests_mock.response import _MatcherResponse
+
+try:
+    import purl
+    purl_types = (purl.URL,)
+except ImportError:
+    purl = None
+    purl_types = ()
 
 ANY = object()
 
 
-class _RequestObjectProxy(object):
-    """A wrapper around a requests.Request that gives some extra information.
-
-    This will be important both for matching and so that when it's save into
-    the request_history users will be able to access these properties.
-    """
-
-    def __init__(self, request):
-        self._request = request
-        self._matcher = None
-        self._url_parts_ = None
-        self._qs = None
-
-    def __getattr__(self, name):
-        return getattr(self._request, name)
-
-    @property
-    def _url_parts(self):
-        if self._url_parts_ is None:
-            self._url_parts_ = urlparse.urlparse(self._request.url.lower())
-
-        return self._url_parts_
-
-    @property
-    def scheme(self):
-        return self._url_parts.scheme
-
-    @property
-    def netloc(self):
-        return self._url_parts.netloc
-
-    @property
-    def path(self):
-        return self._url_parts.path
-
-    @property
-    def query(self):
-        return self._url_parts.query
-
-    @property
-    def qs(self):
-        if self._qs is None:
-            self._qs = urlparse.parse_qs(self.query)
-
-        return self._qs
-
-    @classmethod
-    def _create(cls, *args, **kwargs):
-        return cls(requests.Request(*args, **kwargs).prepare())
-
-    @property
-    def text(self):
-        body = self.body
-
-        if isinstance(body, six.binary_type):
-            body = body.decode('utf-8')
-
-        return body
-
-    def json(self, **kwargs):
-        return json.loads(self.text, **kwargs)
-
-    @property
-    def matcher(self):
-        """The matcher that this request was handled by.
-
-        The matcher object is handled by a weakref. It will return the matcher
-        object if it is still available - so if the mock is still in place. If
-        the matcher is not available it will return None.
-        """
-        return self._matcher()
-
-
 class _RequestHistoryTracker(object):
 
     def __init__(self):
@@ -117,15 +50,30 @@ class _RequestHistoryTracker(object):
     def called(self):
         return self.call_count > 0
 
+    @property
+    def called_once(self):
+        return self.call_count == 1
+
     @property
     def call_count(self):
         return len(self.request_history)
 
 
+class _RunRealHTTP(Exception):
+    """A fake exception to jump out of mocking and allow a real request.
+
+    This exception is caught at the mocker level and allows it to execute this
+    request through the real requests mechanism rather than the mocker.
+
+    It should never be exposed to a user.
+    """
+
+
 class _Matcher(_RequestHistoryTracker):
     """Contains all the information about a provided URL to match."""
 
-    def __init__(self, method, url, responses, complete_qs, request_headers):
+    def __init__(self, method, url, responses, complete_qs, request_headers,
+                 additional_matcher, real_http, case_sensitive):
         """
         :param bool complete_qs: Match the entire query string. By default URLs
             match if all the provided matcher query arguments are matched and
@@ -133,15 +81,42 @@ class _Matcher(_RequestHistoryTracker):
             require that the entire query string needs to match.
         """
         super(_Matcher, self).__init__()
+
         self._method = method
         self._url = url
-        try:
-            self._url_parts = urlparse.urlparse(url.lower())
-        except:
-            self._url_parts = None
         self._responses = responses
         self._complete_qs = complete_qs
         self._request_headers = request_headers
+        self._real_http = real_http
+        self._additional_matcher = additional_matcher
+
+        # url can be a regex object or ANY so don't always run urlparse
+        if isinstance(url, six.string_types):
+            url_parts = urlparse.urlparse(url)
+            self._scheme = url_parts.scheme.lower()
+            self._netloc = url_parts.netloc.lower()
+            self._path = url_parts.path or '/'
+            self._query = url_parts.query
+
+            if not case_sensitive:
+                self._path = self._path.lower()
+                self._query = self._query.lower()
+
+        elif isinstance(url, purl_types):
+            self._scheme = url.scheme()
+            self._netloc = url.netloc()
+            self._path = url.path()
+            self._query = url.query()
+
+            if not case_sensitive:
+                self._path = self._path.lower()
+                self._query = self._query.lower()
+
+        else:
+            self._scheme = None
+            self._netloc = None
+            self._path = None
+            self._query = None
 
     def _match_method(self, request):
         if self._method is ANY:
@@ -160,18 +135,20 @@ class _Matcher(_RequestHistoryTracker):
         if hasattr(self._url, 'search'):
             return self._url.search(request.url) is not None
 
-        if self._url_parts.scheme and request.scheme != self._url_parts.scheme:
+        # scheme is always matched case insensitive
+        if self._scheme and request.scheme.lower() != self._scheme:
             return False
 
-        if self._url_parts.netloc and request.netloc != self._url_parts.netloc:
+        # netloc is always matched case insensitive
+        if self._netloc and request.netloc.lower() != self._netloc:
             return False
 
-        if (request.path or '/') != (self._url_parts.path or '/'):
+        if (request.path or '/') != self._path:
             return False
 
         # construct our own qs structure as we remove items from it below
         request_qs = urlparse.parse_qs(request.query)
-        matcher_qs = urlparse.parse_qs(self._url_parts.query)
+        matcher_qs = urlparse.parse_qs(self._query)
 
         for k, vals in six.iteritems(matcher_qs):
             for v in vals:
@@ -210,15 +187,30 @@ class _Matcher(_RequestHistoryTracker):
 
         return True
 
+    def _match_additional(self, request):
+        if callable(self._additional_matcher):
+            return self._additional_matcher(request)
+
+        if self._additional_matcher is not None:
+            raise TypeError("Unexpected format of additional matcher.")
+
+        return True
+
     def _match(self, request):
         return (self._match_method(request) and
                 self._match_url(request) and
-                self._match_headers(request))
+                self._match_headers(request) and
+                self._match_additional(request))
 
     def __call__(self, request):
         if not self._match(request):
             return None
 
+        # doing this before _add_to_history means real requests are not stored
+        # in the request history. I'm not sure what is better here.
+        if self._real_http:
+            raise _RunRealHTTP()
+
         if len(self._responses) > 1:
             response_matcher = self._responses.pop(0)
         else:
@@ -232,12 +224,15 @@ class Adapter(BaseAdapter, _RequestHistoryTracker):
     """A fake adapter than can return predefined responses.
 
     """
-    def __init__(self):
+    def __init__(self, case_sensitive=False):
         super(Adapter, self).__init__()
+        self._case_sensitive = case_sensitive
         self._matchers = []
 
     def send(self, request, **kwargs):
-        request = _RequestObjectProxy(request)
+        request = _RequestObjectProxy(request,
+                                      case_sensitive=self._case_sensitive,
+                                      **kwargs)
         self._add_to_history(request)
 
         for matcher in reversed(self._matchers):
@@ -264,20 +259,34 @@ class Adapter(BaseAdapter, _RequestHistoryTracker):
         :param str url: The URL to match.
         """
         complete_qs = kwargs.pop('complete_qs', False)
+        additional_matcher = kwargs.pop('additional_matcher', None)
         request_headers = kwargs.pop('request_headers', {})
+        real_http = kwargs.pop('_real_http', False)
 
         if response_list and kwargs:
             raise RuntimeError('You should specify either a list of '
                                'responses OR response kwargs. Not both.')
+        elif real_http and (response_list or kwargs):
+            raise RuntimeError('You should specify either response data '
+                               'OR real_http. Not both.')
         elif not response_list:
-            response_list = [kwargs]
+            response_list = [] if real_http else [kwargs]
 
-        responses = [response._MatcherResponse(**k) for k in response_list]
+        # NOTE(jamielennox): case_sensitive is not present as a kwarg because i
+        # think there would be an edge case where the adapter and register_uri
+        # had different values.
+        # Ideally case_sensitive would be a value passed to match() however
+        # this would change the contract of matchers so we pass ito to the
+        # proxy and the matcher separately.
+        responses = [_MatcherResponse(**k) for k in response_list]
         matcher = _Matcher(method,
                            url,
                            responses,
+                           case_sensitive=self._case_sensitive,
                            complete_qs=complete_qs,
-                           request_headers=request_headers)
+                           additional_matcher=additional_matcher,
+                           request_headers=request_headers,
+                           real_http=real_http)
         self.add_matcher(matcher)
         return matcher
 
